/**
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 -
 - JASS: Java Advanced tranSaction Support
 -
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 -
 - This module was originally developed by 
 -
 -    LSD (Distributed Systems Lab, http://lsd.ls.fi.upm.es/lsd/lsd.htm)
 -    at Universidad Politecnica de Madrid (UPM) as an ObjectWeb Consortium
 -    (http://www.objectweb.org) project.
 -
 -    This project has been partially funded by the European Commission under
 -    the IST programme of V FP grant IST-2001-37126 and by the Spanish
 -    Ministry of Science & Technology (MCyT) grants TIC2002-10376-E and
 -    TIC2001-1586-C03-02
 -
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 -  The original code and portions created by LSD are 
 -  Copyright (c) 2004 LSD (UPM)  
 -  All rights reserved.
 -  
 - Redistribution and use in source and binary forms, with or without 
 - modification, are permitted provided that the following conditions are met:
 -
 - -Redistributions of source code must retain the above copyright notice, this
 - list of conditions and the following disclaimer. 
 -
 - -Redistributions in binary form must reproduce the above copyright notice, 
 - this list of conditions and the following disclaimer in the documentation 
 - and/or other materials provided with the distribution. 
 -
 - THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 - AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 - IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 - ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 - LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 - CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 - SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 - INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 - CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 - ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 - POSSIBILITY OF SUCH DAMAGE.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 -
 -  Author: Francisco Perez-Sorrosal (fpsorrosal@fi.upm.es)
 -
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

package org.objectweb.jass.as;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;

import javax.activity.ActionErrorException;
import javax.activity.ActionNotFoundException;
import javax.activity.ActivityCoordinator;
import javax.activity.ActivityManager;
import javax.activity.ActivityNotProcessedException;
import javax.activity.ActivityPendingException;
import javax.activity.CompletionStatus;
import javax.activity.ContextPendingException;
import javax.activity.CoordinationInformation;
import javax.activity.GlobalId;
import javax.activity.InvalidActivityException;
import javax.activity.InvalidStateException;
import javax.activity.NotOriginatorException;
import javax.activity.Outcome;
import javax.activity.Signal;
import javax.activity.SignalSetInactiveException;
import javax.activity.SignalSetUnknownException;
import javax.activity.Status;
import javax.activity.SystemException;
import javax.activity.coordination.Action;
import javax.activity.coordination.SignalSet;

import org.apache.log4j.Logger;
import org.objectweb.jass.as.exceptions.SignalSetNotRegisteredException;
import org.objectweb.jass.as.util.SignalSetInfo;
import org.objectweb.jass.as.util.Timeout;
import org.objectweb.jass.as.util.TimeoutFactory;
import org.objectweb.jass.as.util.TimeoutTarget;

/**
 * Each object of this class represents an activity and its coordinator. 
 * Implements the AS ActivityCoordinator interface in order to be able to 
 * broadcast signals to registered actions. TimeoutTarget is implemented in 
 * order to be warned of timeout.
 * @author fran
 * Date: Feb 16, 2004
 * org.objectweb.jass.asActivityImpl.java
 */
public class ActivityImpl implements ActivityCoordinator, TimeoutTarget, Serializable {

   // Constants -----------------------------------------------------

   // Attributes ----------------------------------------------------

   private static Logger log = Logger.getLogger(ActivityImpl.class);
   // Unique activity identifier
   private ActivityIdImpl activityId;
   // Indicates the superior ActivityCoordinator of the activity
   private ActivityImpl superior = null;
   // Contains the children of the activity
   private LinkedList children = new LinkedList();
   // Activity current status
   private int status = Status.StatusNoActivity;
   // Activity completion status
   private int completionStatus = CompletionStatus.CompletionStatusFail;
   // The time the activity was started
   private long start;
   // Activity timeout
   private transient Timeout timeout = null; // I use an modified implementation of the TimeoutFactory
   private boolean timedOut = false;
   // Contains objects representing information about the registered
   // signalSets. The keys are the signalSetNames
   private HashMap signalSets = new HashMap();
   // The name of the completion SignalSet
   private String completionSSName = null;
   // A reference to UserActivity implementation
   private transient ActivityService manager = null; // DISALLOW SERIALIZATION
   // Mutex for thread-safety. This should only be changed in the lock()
   // and unlock() methods
   private boolean locked = false;
   // Flags that the activity has finished
   private boolean done = false;
   private transient HashSet threads = new HashSet(); // DISALLOW SERIALIZATION

   // Constructors ---------------------------------------------------------------

   public ActivityImpl(
      ActivityService activityService,
      ActivityIdImpl activityId,
      ActivityImpl superior,
      int timeout) {
      
      log.debug("STARTING ACTIVITY " + activityId.print());
      this.manager = activityService;
      this.activityId = activityId;
      this.superior = superior;
      // Set the starting time of the currentActivity and
      // the timeout in milliseconds (timeout parameter in secs * 1000)
      start = System.currentTimeMillis();
      // We only set the timeout when its distinct from zero
      if (timeout != 0) {
         this.timeout = TimeoutFactory.createTimeout(start + (timeout * 1000), this);
         log.debug("\tTimeout = " + timeout);
      } else
         log.debug("\tTimeout NOT set to this activity");
      // Change to the right status
      status = Status.StatusActive;
      log.debug(activityId.print() + " STARTED!!!");
   }

   // Public -----------------------------------------------------------------

   // ActivityCoordinator implementation -------------------------------------

   /**
    * Establishes an interest relationship between the specified Action and
    * SignalSet for the Activity represented by the target ActivityCoordinator.
    * When a Signal which is a member of the SignalSet is sent, the
    * processSignal method of the Action will be invoked with that  Signal. If
    * multiple Actions are registered, then priority may be used to place an
    * order on how they will be invoked when Signals are sent: higher priority
    * Actions will occur first in the Action list, and hence be invoked before
    * other, lower priority, Actions. The priority value must be a positive
    * value; a value of zero means that the Activity Service implementation is
    * free to place the Action at any point in the Action list. If the
    * specified Action is registered multiple times for the same SignalSet
    * then it will be invoked multiple times with the Signals from that
    * SignalSet.
    */
   public void addAction(Action action, String signalSetName, int priority)
      throws SignalSetUnknownException, SystemException, IllegalStateException {

      SignalSetInfo ssInfo = null;
      SignalSet signalSet = null;

      log.debug(activityId.print());
      // If the SignalSet is not registered in the activity (signalSets HashMap)...
      if (!signalSets.containsKey(signalSetName)) {

         log.debug("\tSignalSet " + signalSetName + " not yet registered.");
         log.debug("\tRegistering SignalSet and adding action...");

         try {
            // ... first, we obtain an instance and associate with its name...
            signalSet = manager.getService().getSignalSet(signalSetName);
            signalSet.setActivityCoordinator(this);
            ssInfo = new SignalSetInfo(signalSet);
            ssInfo.addAction(action, priority, false);
            signalSets.put(signalSetName, ssInfo);
            log.debug("\tSignalSet " + signalSetName + " registered");
         } catch (SignalSetUnknownException e) {
            log.warn("Null SignalSet");
            e.printStackTrace();
         } catch (Exception e) {
            e.printStackTrace();
         }
      } else { // The signalSet already exists in the HashMap
         log.debug("\tSignalSet " + signalSetName + " already registered. Adding action...");
         ssInfo = (SignalSetInfo) signalSets.get(signalSetName);
         ssInfo.addAction(action, priority, false);
         signalSets.put(signalSetName, ssInfo);
      }
      log.debug("\tAction added!!!");
      log.debug("\tActions registered = " + ssInfo.getNumberRegisteredActions());
   }

   /**
    * Establishes an interest relationship between the specified
    * Action and all SignalSets used by the Activity represented
    * by the target ActivityCoordinator. When a Signal which is
    * a member of any of the SignalSets is sent, the processSignal
    * method of the Action will be invoked with that Signal.
    * If multiple Actions are registered, then priority may be used
    * to place an order on how they will be invoked when Signals are
    * sent: higher priority Actions will occur first in the Action
    * list, and hence be invoked before other, lower priority, Actions.
    * The priority value must be a positive value; a value of zero
    * means that the Activity Service implementation is free to place
    * the Action at any point in the Action list.
    * If the specified Action is registered multiple times with the
    * target ActivityCoordinator then it will be invoked multiple times
    * with each Signal.
    */
   public void addGlobalAction(Action action, int priority)
      throws SystemException, java.lang.IllegalStateException {

      // First, we get all registered signal sets...
      Set keys = signalSets.keySet();
      Iterator it = keys.iterator();
      // ...and then we iterate on them adding the action.
      while (it.hasNext()) {
         SignalSetInfo ssInfo = (SignalSetInfo) it.next();
         ssInfo.addAction(action, priority, true);
      }
   }

   /**
    * Removes the interest relationship between the specified Action
    * and SignalSet for the Activity represented by the target
    * ActivityCoordinator. If signalSetName is an empty String then
    * the interest relationship is removed between the specified Action
    * and all SignalSets in the target Activity. This operation has no
    * effect on any global actions.
    * If the specified Action was registered multiple times with the
    * target ActivityCoordinator then only a single instance of this Action
    * is removed by each call to this method.
    */
   public void removeAction(Action action, String signalSetName)
      throws ActionNotFoundException, SystemException, IllegalStateException {

      boolean result = false;

      if ((status == Status.StatusCompleting)
         || (status == Status.StatusCompleted)
         || (status == Status.StatusCompletingHeuristic))
         throw new IllegalStateException();

      if (!signalSets.containsKey(signalSetName))
         throw new SystemException("SignalSet " + signalSetName + " not registered");
      result = ((SignalSetInfo) signalSets.get(signalSetName)).removeAction(action, false);

      if (!result)
         throw new ActionNotFoundException();
   }

   /**
    * Removes the interest relationship between the specified Action and all
    * SignalSets  for the Activity represented by the target ActivityCoordinator.
    * This operation has no effect on any non-global actions. If the specified 
    * Action was registered multiple times with the target ActivityCoordinator
    * then only a single instance of this Action is removed by each call to 
    * this method.
    */
   public void removeGlobalAction(Action action)
      throws ActionNotFoundException, SystemException, IllegalStateException {

      if ((status == Status.StatusCompleting)
         || (status == Status.StatusCompleted)
         || (status == Status.StatusCompletingHeuristic))
         throw new IllegalStateException();

      // First, we get all registered signal sets...
      Set keys = signalSets.keySet();
      Iterator it = keys.iterator();
      // ...and then we iterate on them removing the action.
      while (it.hasNext()) {
         SignalSetInfo ssInfo = (SignalSetInfo) it.next();
         ssInfo.removeAction(action, true);
      }
   }

   /** 
    * Returns all the Actions that have been registered with
    * an interest in the specified signalSetName. This method may
    * be used by a HLS that has determined that an Activity needs
    * to be persisted and needs to log the Actions registered in the Activity.	 
    */
   public Action[] getActions(java.lang.String signalSetName)
      throws SignalSetUnknownException, SystemException {

      if (!signalSets.containsKey(signalSetName))
         throw new SignalSetUnknownException();

      return ((SignalSetInfo) signalSets.get(signalSetName)).getActions();
   }

   /**
    * Returns the number of Actions, including global Actions,
    * registered with the target ActivityCoordinator with an
    * interest in the specified SignalSet.
    */
   public int getNumberRegisteredActions(String signalSetName)
      throws SignalSetUnknownException, SystemException {

      if (!signalSets.containsKey(signalSetName))
         throw new SignalSetUnknownException();

      return ((SignalSetInfo) signalSets.get(signalSetName)).getNumberRegisteredActions();
   }

   /**
    * Sets the name of the SignalSet that should be used for the distribution 
    * of completion signals when the current Activity completes.
    * This method can be called multiple times during the lifetime of an 
    * Activity, before the Activity starts completion processing.
    * The last value specified before completion processing starts is used
    * during completion processing.
    */
   public void setCompletionSignalSetName(String signalSetName)
      throws SignalSetUnknownException, SystemException, IllegalStateException {

      completionSSName = signalSetName;
   }

   /**
    * Returns the name of the SignalSet, if any, that will be used for the
    * distribution of completion signals when the current Activity completes.
    */
   public String getCompletionSignalSetName() throws SystemException {

      return completionSSName;
   }

   /**
    * Returns the parent ActivityCoordinator or null if the target 
    * ActivityCoordinator represents a top-level Activity.
    */
   public ActivityCoordinator getParent() throws SystemException {

      return superior;
   }

   /**
    * Returns the GlobalId of the Activity represented by the target ActivityCoordinator.
    */
   public GlobalId getGlobalId() throws SystemException {

      return activityId;
   }

   /**
    * Returns the Status of the Activity represented by the target ActivityCoordinator.
    */
   public int getStatus() throws SystemException {

      return status;
   }

   /**
    * Returns the Status of the Activity represented by the target ActivityCoordinator's
    * parent. If the target ActivityCoordinator represents a top-level
    * Activity, then the Status of this top-level Activity is returned.
    */
   public int getParentStatus() throws SystemException {

      if (superior != null)
         return superior.getStatus();
      else
         return status;
   }

   /**
    * Returns a printable string describing the Activity represented by the target
    * ActivityCoordinator.
    */
   public String getName() throws SystemException {
      return activityId.print();
   }

   /**
    * Returns true if the specified coord represents the same Activity as the
    * target ActivityCoordinator.
    */
   public boolean isSameActivity(ActivityCoordinator coord) throws SystemException {

      if (coord != null && coord instanceof ActivityImpl)
         return this.activityId.equals(coord.getGlobalId());

      return false;
   }

   /**
    * Causes the Activity associated with the target ActivityCoordinator to
    * complete with the specified CompletionStatus. The Activity service sets
    * the Status to StatusCompleting before asking any completion SignalSet
    * to start producing Signals.
    * If there are any child active or suspended Activities or transactions
    * and the CompletionStatus is CompletionStatusFail or
    * CompletionStatusFailOnly then those child Activites will have their
    * CompletionStatuses set to CompletionStatusFailOnly and any child
    * transaction will be called to setRollbackOnly.
    * The UserActivity interface should be used in preference to the
    * ActivityCoordinator for completing Activities in most circumstances.
    * This method is provided primarily for use after recovery when the
    * Activity to be completed is not associated with the calling thread.
    */
   public Outcome completeActivity(int completionStatus)
      throws
         ActivityPendingException,
         ContextPendingException,
         NotOriginatorException,
         InvalidStateException,
         ActivityNotProcessedException,
         SystemException {
      // Set the completion status of this instance and...
      this.completionStatus = completionStatus;
      // ...delegate on the implementation dependant method.
      return completeActivity();
   }

   /**
    * TODO NOT YET IMPLEMENTED
    */
   public Outcome heuristicComplete(int completionStatus)
      throws
         ActivityPendingException,
         ContextPendingException,
         InvalidStateException,
         ActivityNotProcessedException,
         SystemException {
      return null;
   }

   /**
    * Causes the SignalSet specified by signalSetName to start producing
    * signals for all registered Actions at times other than during completion.
    * Pre-defined SignalSets, such as Synchronization, cannot be requested to
    * produce signals via this method.
    */
   public Outcome processSignalSet(String signalSetName, int completionStatus)
      throws
         SignalSetUnknownException,
         ActivityNotProcessedException,
         InvalidActivityException,
         SystemException {

      SignalSet processingSS = null;
      Signal signal = null;
      Outcome out = null;
      Action[] actions = null;

      // If the activity is finished, throw an exception
      if (done)
         throw new InvalidActivityException();
      // If the SignalSet is not registered, throw an exception
      if (!signalSets.containsKey(signalSetName))
         throw new SignalSetUnknownException();

      // First, get the SignalSet to process...
      try {
         processingSS = getSignalSet(signalSetName);
         processingSS.setCompletionStatus(completionStatus, status);
      } catch (SignalSetNotRegisteredException e) {
         status = Status.StatusError;
         throw new ActivityNotProcessedException();
      }

      log.debug("Processing " + signalSetName);

      if ((actions = getActions(signalSetName)) == null)
         log.debug("\tNO actions registered with " + signalSetName);
      else { // If there are actions to process...
         try {
            // ...then deliver the signals...
            deliverSignals(processingSS, actions);
            // ...and finally, get the SignalSet outcome.
            out = processingSS.getOutcome();
         } catch (Exception e) {
            e.printStackTrace();
         }
      }

      return out;
   }

   // Implements TimeoutTarget -----------------------------------------------

   /**
    * Called when our timeout expires.
    */
   public void timedOut(Timeout timeout) {
      try {
         lock();
         log.warn("Activity " + activityId.print() + " TIMED OUT !!!!!");
         log.warn(" Status = " + getStringStatus(status));
         if (this.timeout == null)
            return; // Don't race with timeout cancellation.
         this.timeout = null;
         timedOut = true;
         switch (status) {
            case Status.StatusCompleting :
            case Status.StatusCompleted :
               // Give a chance to finish
               return;
            case Status.StatusActive :
            case Status.StatusError :
               completionStatus = CompletionStatus.CompletionStatusFailOnly;
               // We mark the completion status of the children of this activity 
               // as CompletionStatusFailOnly                                    
               markSubActivitiesWithFailOnly();
               return;
            default :
               log.warn("Unknown status at timeout, Activity = " + toString());
               status = Status.StatusUnknown;
               return;
         }
      } finally {
         unlock();
      }
   }

   // Implements Serialization ------------------------------------------------
   
   private void writeObject(ObjectOutputStream s) throws IOException {
    s.defaultWriteObject();
    // Get the current remaining time for the activity
    if(timeout != null) {
    		long remainingTime = timeout.getRemainingTime();
    		s.writeLong(remainingTime);
    } else {
    		s.writeLong(-1);
   	}
    
   }
   
   private void readObject(ObjectInputStream s) throws IOException  {
   	try {
   		s.defaultReadObject();
   	} catch (ClassNotFoundException e) {
   		e.printStackTrace();
   	}
   	long remainingTime = s.readLong();
   	if(remainingTime != -1) {
   		// Set the remaining time to the activity
   		timeout = TimeoutFactory.createTimeout(remainingTime, this);
   	}
   	// Set the local logger
   	log = Logger.getLogger(ActivityImpl.class);
   	// Create the thread set
   	threads = new HashSet();
   }

   // My Public Methods ------------------------------------------------------

   /**
    * Associates the calling thread with this activity.
    */
   public void associateCurrentThread() {

      /*log.warn(
      	"Associating thread "
      		+ getThreadName()
      		+ " with activity "
      		+ activityId.print());*/
      threads.add(Thread.currentThread());
      //log.warn("Associated threads = " + threads.size());
   }

   /**
    * Disassociates the calling thread from this activity.
    */
   public void disassociateCurrentThread() {

      /*log.warn(
      	"Disassociating thread "
      		+ getThreadName()
      		+ " from activity "
      		+ activityId.print());*/
      threads.remove(Thread.currentThread());
      //log.warn("Associated threads = " + threads.size());
      Thread.interrupted();
   }

   /**
    * Reasociates the local ActivityService with the recreated activity
    *
    * @param am the activity manager interface representing the Activity 
    * Service implementation.
    */
   public void associateActivityManager(ActivityManager am) {
   	this.manager = (ActivityService) am;
   }
   
   /**
    *  Return the string representation of the activity status.
    *
    * @param status the integer status to convert.
    * @return the string representation of the status.  
    */
   public String getStringStatus(int status) {

      switch (status) {
         case Status.StatusActive :
            return "STATUS_ACTIVE";
         case Status.StatusCompleted :
            return "STATUS_COMPLETED";
         case Status.StatusCompleting :
            return "STATUS_COMPLETING";
         case Status.StatusCompletingHeuristic :
            return "STATUS_COMPLETING_HEURISTIC";
         case Status.StatusError :
            return "STATUS_ERROR";
         case Status.StatusNoActivity :
            return "STATUS_NO_ACTIVITY";
         case Status.StatusUnknown :
            return "STATUS_UNKNOWN";

         default :
            return "STATUS_UNKNOWN(" + status + ")";
      }
   }
   /**
    * Return the string representation of the activity completion status.
    * 
    * @param completionStatus the integer completion status to convert.
    * @return the string representation of the completion status.
    */
   public String getStringCompletionStatus(int completionStatus) {

      switch (completionStatus) {
         case CompletionStatus.CompletionStatusFail :
            return "COMPLETION_STATUS_FAIL";
         case CompletionStatus.CompletionStatusFailOnly :
            return "COMPLETION_STATUS_FAIL_ONLY";
         case CompletionStatus.CompletionStatusSuccess :
            return "COMPLETION_STATUS_SUCCESS";

         default :
            return "COMPLETION_STATUS_UNKNOWN(" + completionStatus + ")";
      }
   }
   
   // My Protected ---------------------------------------------------------------

   /**
    * Protected method to complete the activity.
    */
   protected Outcome completeActivity()
      throws
         ActivityPendingException,
         ContextPendingException,
         NotOriginatorException,
         InvalidStateException,
         ActivityNotProcessedException,
         SystemException {

      SignalSet processingSS = null;
      Outcome out = null;
      Action[] actions = null;

      log.debug("COMPLETING ACTIVITY " + activityId.print());

      if (threads.size() > 1) {
         log.debug("Threads size  - >" + threads.size());
         throw new ActivityPendingException();
         
      }
         

      // If the completion status for this acitiviti is fail or fail only, we 
      // must mark any encompassed activity with FAIL ONLY
      // and make orphan the activity children  
      if (((completionStatus == CompletionStatus.CompletionStatusFail)
         || (completionStatus == CompletionStatus.CompletionStatusFailOnly))
         && (children.size() > 0)) {
         markSubActivitiesWithFailOnly();
         makeOrphan();
      }

      status = Status.StatusCompleting;

      if (timedOut) // If timeout has expired, complete with FAIL
         completionStatus = CompletionStatus.CompletionStatusFail;
      if (timeout != null)
         timeout.cancel();

      log.debug("\tWITH " + getStringCompletionStatus(completionStatus));

      if (completionSSName == null) {
         // It may be not have been set yet by the HLS.
         completionSSName = manager.getService().getCompletionSignalSetName();
         // If still it is null, then return a null outcome.
         if (completionSSName == null) {
            log.debug("\tThere is no completion SignalSet to process.");
            // We finish the instance
            log.debug(activityId.print() + " COMPLETED!!!");
            instanceDone();
            return out;
         }
      }

      // The completion SignalSet may or not have been registered previously
      try {
         if (signalSets.containsKey(completionSSName)) {
            if (log.isDebugEnabled()) log.debug("SignalSet " + completionSSName + " already registered");
            processingSS = getSignalSet(completionSSName);
            actions = getActions(completionSSName);
         } else {
            if (log.isDebugEnabled()) log.debug("SignalSet " + completionSSName + " not registered");
            processingSS = manager.getService().getSignalSet(completionSSName);
            processingSS.setActivityCoordinator(this);
         }
      } catch (Exception e) {
         // SignalSetUnknownException, SignalSetNotRegisteredException
         // NoImplementException, SignalSetActiveException
         status = Status.StatusError;
         throw new ActivityNotProcessedException();
      }

      log.debug("\tProcessing Completion SignalSet -> " + completionSSName);
      processingSS.setCompletionStatus(completionStatus, status);

      try {
         if (actions == null) {
            log.debug("\tNO actions registered");
         } else {
            deliverSignals(processingSS, actions);
            // WARNING!!! We can't modify a CompletionStatusFailOnly completionStatus
            int signalSetCompStat = processingSS.getCompletionStatus();
            if ((signalSetCompStat != CompletionStatus.CompletionStatusFailOnly)
               && (completionStatus == CompletionStatus.CompletionStatusFailOnly))
               throw new InvalidStateException();
            completionStatus = signalSetCompStat;
            //	Finally, we obtain the SignalSet outcome and CompletionStatus
            out = processingSS.getOutcome();
         }
         processingSS.destroy();
         log.debug("\tCompletion SignalSet processing completed");
         log.debug("\tFinal CompletionStatus = " + getStringCompletionStatus(completionStatus));
         status = Status.StatusCompleted;
      } catch (Exception e) {
         e.printStackTrace();
      } finally {
         // We finish the instance
         log.debug(activityId.print() + " COMPLETED!!!");
         instanceDone();
      }
      return out;
   }

   /**
    * Delivers the SignalSet signals to registered actions. 
    * @param signalSet the triggered SignalSet.
    * @param actions the actions registered with the SignalSet.
    * @throws ActionErrorException
    */
   private void deliverSignals(SignalSet signalSet, Action[] actions) throws ActionErrorException {

      Signal signal = null;
      CoordinationInformation info = null;
      Outcome out = null;

      while ((signal = signalSet.getSignal()) != null) {
         int i = 0;
         boolean nextSignal = false;
         log.debug("\tSending " + signal.getName() + " signal...");
         log.debug("\t# of registered actions = " + actions.length);
         while ((i < actions.length) && (nextSignal == false)) {
            if (actions[i] != null) {
               if (log.isDebugEnabled()) log.debug("\tSending number: " + i);
               out = actions[i].processSignal(signal);
               // Get and check the coordination information						
               try {
                  info = signalSet.setResponse(out);
               } catch (SignalSetInactiveException e) {
                  e.printStackTrace();
               }
               if ((info != null) && (info.isActionInterested() == false)) {
                  // We must remove the action in order to not process
                  // it in other processings while this activity still is alive...					
                  SignalSetInfo ssi =
                     ((SignalSetInfo) signalSets.get(signalSet.getSignalSetName()));
                  ssi.removeAction(actions[i], false);
                  // ...and remove action from array for the current processing...
                  actions[i] = null;
               }
               if ((info != null) && (info.useNextSignal() == true)) {
                  nextSignal = true;
                  break; // We take other signal
               }
            }
            i = i + 1;
         }
      }
   }

   /**
    * Sets the completionStatus for this activity.
    */
   protected void setCompletionStatus(int completionStatus) throws InvalidStateException {

      if (this.completionStatus == CompletionStatus.CompletionStatusFailOnly)
         throw new InvalidStateException();
      this.completionStatus = completionStatus;
   }

   /**
    * Gets the completionStatus of this activity.
    */
   protected int getCompletionStatus() {
      return completionStatus;
   }

   /**
    * Adds a child to the children list.
    * @param child - The new child.
    */
   protected void addChild(ActivityImpl child) {
      children.add(child);
   }

   /**
    * Removes a child from the list of children.
    * @param child - The child to remove.
    */
   protected void removeChild(ActivityImpl child) {
      children.remove(child);
   }

   /**
    * Returns an array of activity children.
    * @return - The array with all the children of this activity.
    */
   protected ActivityImpl[] getChilds() {
      ActivityImpl[] array = null;

      if (children.size() > 0)
         array = (ActivityImpl[]) children.toArray(new ActivityImpl[children.size()]);
      return array;
   }

   // My Private -------------------------------------------------------------

   /**
    * Returns a SignalSet from the registered signal sets.
    * @param signalSetName the name of the SignalSet to return.
    * @return the registered SignalSet.
    * @throws SignalSetNotRegisteredException thrown if the requested SignalSet
    * has not been registered.
    */
   private SignalSet getSignalSet(String signalSetName) throws SignalSetNotRegisteredException {

      SignalSetInfo signalSet = (SignalSetInfo) signalSets.get(signalSetName);
      if (signalSet == null)
         throw new SignalSetNotRegisteredException();
      return signalSet.getSignalSet();
   }

   /**
    * Mark any encompassed activity with completion status FAIL ONLY. This is 
    * done either when the timeout expires or when the activity is going to complete 
    * with FAIL/FAIL ONLY.
    */
   private void markSubActivitiesWithFailOnly() {
      ActivityImpl child = null;
      if (children != null) {
         try {
            if (log.isDebugEnabled()) log.debug("Marking with FAIL childs... of " + getName());
            Iterator it = children.iterator();
            while (it.hasNext()) {
               child = (ActivityImpl) it.next();
               child.setCompletionStatus(CompletionStatus.CompletionStatusFailOnly);
               if (log.isDebugEnabled()) log.debug("Child " + child.getName() + " marked!!!");
               child.markSubActivitiesWithFailOnly();
            }
         } catch (InvalidStateException e) {
            if (log.isDebugEnabled()) log.debug("Child " + child + "is already marked as FAIL ONLY");            
         } catch (SystemException e) {
            e.printStackTrace();
         }
      }
   }

   /**
    * Make orphan the activity children
    */
   private void makeOrphan() {
      ActivityImpl child = null;
      if (children != null) {
         if (log.isDebugEnabled()) log.debug("Making orphan children of " + activityId.print());
         Iterator it = children.iterator();
         while (it.hasNext()) {
            child = (ActivityImpl) it.next();
            log.debug("\tChild " + child.activityId.print() + " is orphan now");
            child.superior = null;
         }
      }
   }


   /**
    *  Lock this instance.
    */
   private synchronized void lock() {
      if (done)
         throw new IllegalStateException("Activity has terminated");
      if (locked) {
         log.warn("Lock contention, activity=" + toString());
         //DEBUG Thread.currentThread().dumpStack();
         while (locked) {
            try {
               // Wakeup happens when:
               // - notify() is called from unlock()
               // - notifyAll is called from instanceDone()
               wait();
            } catch (InterruptedException ex) {
               // ignore
            }
            if (done)
               throw new IllegalStateException("Activity has now terminated");
         }
      }
      locked = true;
   }

   /**
    *  Unlock this instance.
    */
   private synchronized void unlock() {
      if (!locked) {
         log.warn(
            "Unlocking, but not locked, activity=" + toString(),
            new Throwable("[Stack trace]"));
      }
      locked = false;
      notify();
   }

   /**
    * Mark this activity as non-existing.
    */
   private synchronized void instanceDone() {
      // If activity is nested, remove the parent association
      if (superior != null)
         superior.removeChild(this);
      // Release this activity from the table for GC
      manager.releaseActivityImpl(this);
      // Set the status to StatusNoActivity
      status = Status.StatusNoActivity;
      // Notify all threads waiting for the lock
      notifyAll();
      // This instance is finished
      done = true;
   }

   /**
    * Returns the name of current thread.
    * @return the current thread name.
    */
   private String getThreadName() {
      return Thread.currentThread().getName();
   }

   // Inner classes -----------------------------------------------------------

   // Debug -------------------------------------------------------------------

   // End Debug ---------------------------------------------------------------
}